var
  CRC_TBL16: array[0..255] of word;

procedure FillTable(const data: array of word);
var
  i: integer;
begin
  for i := 0 to high(data) do
    CRC_TBL16[i] := data[i];
end;

procedure OnLoad();
begin
  FillTable([ // CRC-16, Poly:$1021, RefInOut:False
$0000, $1021, $2042, $3063, $4084, $50A5, $60C6, $70E7, $8108, $9129, $A14A, $B16B, $C18C, $D1AD, $E1CE, $F1EF, 
$1231, $0210, $3273, $2252, $52B5, $4294, $72F7, $62D6, $9339, $8318, $B37B, $A35A, $D3BD, $C39C, $F3FF, $E3DE, 
$2462, $3443, $0420, $1401, $64E6, $74C7, $44A4, $5485, $A56A, $B54B, $8528, $9509, $E5EE, $F5CF, $C5AC, $D58D, 
$3653, $2672, $1611, $0630, $76D7, $66F6, $5695, $46B4, $B75B, $A77A, $9719, $8738, $F7DF, $E7FE, $D79D, $C7BC, 
$48C4, $58E5, $6886, $78A7, $0840, $1861, $2802, $3823, $C9CC, $D9ED, $E98E, $F9AF, $8948, $9969, $A90A, $B92B, 
$5AF5, $4AD4, $7AB7, $6A96, $1A71, $0A50, $3A33, $2A12, $DBFD, $CBDC, $FBBF, $EB9E, $9B79, $8B58, $BB3B, $AB1A, 
$6CA6, $7C87, $4CE4, $5CC5, $2C22, $3C03, $0C60, $1C41, $EDAE, $FD8F, $CDEC, $DDCD, $AD2A, $BD0B, $8D68, $9D49, 
$7E97, $6EB6, $5ED5, $4EF4, $3E13, $2E32, $1E51, $0E70, $FF9F, $EFBE, $DFDD, $CFFC, $BF1B, $AF3A, $9F59, $8F78, 
$9188, $81A9, $B1CA, $A1EB, $D10C, $C12D, $F14E, $E16F, $1080, $00A1, $30C2, $20E3, $5004, $4025, $7046, $6067, 
$83B9, $9398, $A3FB, $B3DA, $C33D, $D31C, $E37F, $F35E, $02B1, $1290, $22F3, $32D2, $4235, $5214, $6277, $7256, 
$B5EA, $A5CB, $95A8, $8589, $F56E, $E54F, $D52C, $C50D, $34E2, $24C3, $14A0, $0481, $7466, $6447, $5424, $4405, 
$A7DB, $B7FA, $8799, $97B8, $E75F, $F77E, $C71D, $D73C, $26D3, $36F2, $0691, $16B0, $6657, $7676, $4615, $5634, 
$D94C, $C96D, $F90E, $E92F, $99C8, $89E9, $B98A, $A9AB, $5844, $4865, $7806, $6827, $18C0, $08E1, $3882, $28A3, 
$CB7D, $DB5C, $EB3F, $FB1E, $8BF9, $9BD8, $ABBB, $BB9A, $4A75, $5A54, $6A37, $7A16, $0AF1, $1AD0, $2AB3, $3A92, 
$FD2E, $ED0F, $DD6C, $CD4D, $BDAA, $AD8B, $9DE8, $8DC9, $7C26, $6C07, $5C64, $4C45, $3CA2, $2C83, $1CE0, $0CC1, 
$EF1F, $FF3E, $CF5D, $DF7C, $AF9B, $BFBA, $8FD9, $9FF8, $6E17, $7E36, $4E55, $5E74, $2E93, $3EB2, $0ED1, $1EF0]);
  AddMsg('Plugin Loaded');	
end;

function _CountCrc16(CRC: word; data: byte): word;
begin
  Result := (CRC shl 8) xor CRC_TBL16[(CRC shr 8) xor data];
end;

// CRC-16, Poly:$1021, Ini:$0000, RefInOut:False, Xor:$0000, Check:CRC('123456789')=$31C3
function _CountBlockCrc16(const sData: string): word;
var
  i: Integer;
  CRC: word;
begin
  CRC := $0000;
  for i := 1 to length(sData) do
  begin
    CRC := _CountCrc16(CRC, ord(sData[i]));
  end;
  Result := CRC xor $0000;
end;

const
  S_SOH = $01;
  S_EOT = $04;
  R_ACK = $06;
  R_NAK = $15;
  R_START = $43; // ord('C');
  S_SUB = $1A;

function MakePacket(num: byte; const data_128: string): string;
var
  CRC: word;
begin
  CRC := _CountBlockCrc16(data_128);
  Result := chr(S_SOH) + chr(num) + chr(255 - num) + data_128 + chr(Hi(CRC)) + chr(Lo(CRC));
end;

function WaitForString(timeout_ms: cardinal; ExpResp: byte): string;
var
  sData: string;
begin
  if not Recv(1, timeout_ms, sData) then
    raiseException(erCustomError, 'no answer');
  if (ExpResp <> 0) and (sData <> chr(ExpResp)) then
    raiseException(erCustomError, 'bad answer "' + sData + '"');
  Result := sData;
end;

procedure SendFile(const sData: string);
var
  s: string;
  DPos: Integer;
  gnbl: Integer;
  ato: integer;
begin
  DPos := 0;
  gnbl := 0;
  try
    // wait for start
    WaitForString(4000, R_START);
    inc(gnbl); // 1st nbl should be 1
    while DPos < length(sData) do
    begin
      s := copy(sData, DPos + 1, 128);
      if Length(s) < 128 then
        s := s + StringOfChar(Chr(S_SUB), 128 - Length(s));
	  PurgeRx();
      Send(MakePacket(Lo(gnbl), s));
      ato := 1000;
      if gnbl = 1 then
        ato := 4000; // 4000 - for false start
      s := WaitForString(ato, 0);
      UpdMsg(Format('Rv blk %d answer: %s', [gnbl, s]));
      if s = Chr(R_NAK) then Continue;
      if (gnbl = 1) and (s = Chr(R_START)) then Continue;
      if s <> Chr(R_ACK) then
        raiseException(erCustomError, Format('  "%s"', [s]));
      DPos := DPos + 128;
      inc(gnbl);
    end;
    PurgeRx();
    Send(Chr(S_EOT));
    WaitForString(1000, R_ACK);
    AddMsg(' ');
  except 
  	MsgBox(Format('    %d. %s', [gnbl, ExceptionToString(ExceptionType, ExceptionParam)]), 'script', $0010);
  end;
end;

procedure OnSend();
var
  s: string;
  sData: string;
  i: integer;
begin
  SetV('APTD', '0');
  SetV('APRD', '0');
  if copy(TD, length(TD) - 1, 2) = #13#10 then // "string" mode - delete \r\n from end
    TD := copy(TD, 1, length(TD) - 2);
  if LoadStrFromFile(Trim(TD), sData) then
  	SendFile(sData);
  TD := '';	
end;